import unittest
from unittest.mock import MagicMock, patch

from sklearn.tree import tree

from multi_imbalance.ensemble.mrbbagging import MRBBagging


class TestMRBBagging(unittest.TestCase):
    def test__group_data(self):
        mrbbagging = MRBBagging(1, tree.DecisionTreeClassifier())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", "B", "C"]
        classes, grouped_data = mrbbagging._group_data(x, y)
        self.assertEqual(classes, {'A', 'B', 'C'})
        self.assertEqual(grouped_data, {'C': [[[3, 3, 3], 'C']], 'A': [[[1, 1, 1], 'A']], 'B': [[[2, 2, 2], 'B']]})

    def test__group_data_with_none(self):
        mrbbagging = MRBBagging(1, tree.DecisionTreeClassifier())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", None, "C"]
        with self.assertRaises(AssertionError):
            mrbbagging._group_data(x, y)

    def test_fit_with_invalid_labels(self):
        mrbbagging = MRBBagging(1, MagicMock())
        x = [[1, 1, 1], [2, 2, 2], [3, 3, 3]]
        y = ["A", "B"]
        with self.assertRaises(AssertionError):
            mrbbagging.fit(x, y)

    def test_fit_with_invalid_classifier(self):
        with self.assertRaises(AssertionError):
            MRBBagging(1, None)

    def test_with_invalid_k(self):
        with self.assertRaises(AssertionError):
            MRBBagging(0, tree.DecisionTreeClassifier())


if __name__ == '__main__':
    unittest.main()



"""Collection of imbalanced datasets.

The characteristics of the available datasets are presented in the table below.

 ID    Name             Repository     Class distribution                                            #S       #F


"""

from collections import OrderedDict
import tarfile
from io import BytesIO
from os import makedirs
from os.path import join, isfile

import numpy as np

from sklearn.datasets.base import Bunch

PRE_FILENAME = 'x'
POST_FILENAME = 'data.npz'
DATA_HOME_BASIC = "./../../data/"

MAP_NAME_ID_KEYS = ['1czysty-cut', '2delikatne-cut', '3mocniej-cut', '4delikatne-bezover-cut',
                    'balance-scale', 'cleveland', 'cleveland_v2', 'cmc', 'dermatology',
                    'glass', 'hayes-roth', 'new_ecoli', 'new_led7digit', 'new_vehicle',
                    'new_winequality-red', 'new_yeast', 'thyroid-newthyroid']

MAP_NAME_ID = OrderedDict()
MAP_ID_NAME = OrderedDict()
for v, k in enumerate(MAP_NAME_ID_KEYS):
    MAP_NAME_ID[k] = v + 1
    MAP_ID_NAME[v + 1] = k


def load_datasets(data_home=DATA_HOME_BASIC):
    """Load the benchmark datasets.
        Parameters
        -------
        data_home : Default catalogue in which the data is stored in .tar.gz format.

        Returns
        -------
        datasets : OrderedDict of Bunch object,
            The ordered is defined by ``filter_data``. Each Bunch object ---
            refered as dataset --- have the following attributes:

        dataset.data : ndarray, shape (n_samples, n_features)

        dataset.target : ndarray, shape (n_samples, )

        dataset.DESCR : string
            Description of the each dataset.

    """
    extracted_dir = join(data_home, "extracted")
    datasets = OrderedDict()

    filter_data_ = MAP_NAME_ID.keys()

    for it in filter_data_:
        filename = PRE_FILENAME + str(MAP_NAME_ID[it]) + POST_FILENAME
        filename = join(extracted_dir, filename)
        available = isfile(filename)

        if not available:
            makedirs(extracted_dir, exist_ok=True)
            with open(f'{data_home}data.tar.gz', 'rb') as fin:
                f = BytesIO(fin.read())
            tar = tarfile.open(fileobj=f)
            tar.extractall(path=extracted_dir)

        data = np.load(filename)
        X, y = data['data'], data['label']

        datasets[it] = Bunch(data=X, target=y, DESCR=it)

    return datasets



import glob
import warnings
from collections import OrderedDict, Counter

import numpy as np
import pandas as pd
from pathlib import Path

from IPython.core.display import display
from scipy.io import arff
from sklearn.preprocessing import OneHotEncoder, LabelEncoder, StandardScaler
from sklearn.utils import Bunch


def construct_flat_2pc_df(X, y) -> pd.DataFrame:
    """
    This function takes two dimensional X and one dimensional y arrays, concatenates and returns them as data frame
    Parameters
    ----------
    X two dimensional numpy array
    y one dimensional numpy array with labels
    -------
    Data frame with 3 columns x1 x2 and y and with number of rows equal to number of rows in X
    """

    y = pd.DataFrame({'y': y})
    X_df = pd.DataFrame(data=X, columns=['x1', 'x2'])

    df = pd.concat([X_df, y], axis=1)

    return df


def get_project_root() -> Path:
    """Returns project root folder."""
    return Path(__file__).parent.parent.parent


def preprocess_dataset(path, return_cat_length=False):
    data, meta = arff.loadarff(path)

    df = pd.DataFrame(data)
    y_index = len(df.columns) - 1
    y = df.pop(df.columns[y_index])

    le = LabelEncoder()
    y = le.fit_transform(y)

    categorical_feature_mask = df.dtypes == object

    categorical_cols = df.columns[categorical_feature_mask].tolist()
    non_categorical_cols = df.columns[~categorical_feature_mask].tolist()

    df[categorical_cols] = df[categorical_cols].replace({b'?': np.NaN})
    mode = df.mode().iloc[0]
    mean = df.filter(non_categorical_cols).mean()

    df[categorical_cols] = df.filter(categorical_cols).fillna(mode)
    df[non_categorical_cols] = df.filter(non_categorical_cols).fillna(mean)

    X = pd.get_dummies(df, columns=categorical_cols)
    if return_cat_length:
        return X.to_numpy(), y, len(non_categorical_cols)
    else:
        return X.to_numpy(), y


def load_arff_datasets(return_cat_length=False):
    dataset_paths = glob.glob(f'{get_project_root()}/data/arff/*')

    datasets = OrderedDict()
    for path in sorted(dataset_paths):
        dataset_file = path.split('/')[-1]
        dataset_name = dataset_file.split('.')[0]
        if return_cat_length:
            X, y, cat_length = preprocess_dataset(path, return_cat_length)
            datasets[dataset_name] = Bunch(data=X, target=y, cat_length=cat_length, DESCR=dataset_name)
        else:
            X, y = preprocess_dataset(path, return_cat_length)
            datasets[dataset_name] = Bunch(data=X, target=y, DESCR=dataset_name)

    return datasets


if __name__ == '__main__':
    datasets = load_arff_datasets(return_cat_length=True)